﻿using System;
using System.Drawing;
using System.IO;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Threading.Tasks;
using System.Threading;
using Sig.SignAPI;
using System.Collections.Generic;
using SysDraw = System.Drawing;

namespace SignDocWithPad
{
	public partial class MainForm : Form
	{
		Task? thumbTrackTask;
		ISigning? signing;

		int dpi = 72;

		string signerName = string.Empty;
		string signerReason = string.Empty;
		string signerLocation = string.Empty;
		bool signerStamps = false;

		bool isSigning = false;
		bool isDrawingRectangle = false;
		bool isSignInDocument = false;
		bool isViewThenSign = false;
		Point? mouseDownPosition;
		Point? mouseMovePosition;
		SysDraw.Rectangle? drawnRectangle;



		readonly object documentPictureBoxLock = new object();
		readonly string newLine = Environment.NewLine;

		public MainForm()
		{
			InitializeComponent();

			menuStartWithDoc.Click += OnStartWithDocClicked;
			menuStartSaveDoc.Click += OnStartSaveDocClicked;

			menuDpi.DropDownItemClicked += OnDpiItemClicked;
			menuDpi72.Checked = true;

			menuSignRectangle.Click += OnSignRectangleItemClicked;
			menuSignRectangle.CheckOnClick = true;

			menuSignInDocument.Click += OnSignInDocumentClicked;
			menuSignInDocument.CheckOnClick = true;

			menuViewThenSign.Click += OnViewThenSignClicked;
			menuViewThenSign.CheckOnClick = true;

			menuViewCurrentPage.Click += OnViewCurrentPageClicked;

			menuAskForName.CheckOnClick = true;

			menuPrintTheDocument.Click += PrintTheDocument;

			MainFormModel.log = (string text) => LogToTextBox(text, true);

			MainFormModel.RepeatClicked += () => ClearPictureBox();
			MainFormModel.CancelClicked += () => OnCancelClicked();
			MainFormModel.OkClicked += () => OnOkClicked();
			MainFormModel.OnSignStarted += () => this.BeginInvoke(new Action(AddConfirmRepeatCancelButtons));

			//System.Globalization.CultureInfo frenchCulture = new System.Globalization.CultureInfo("fr");
			//Thread.CurrentThread.CurrentCulture = frenchCulture;
			//Thread.CurrentThread.CurrentUICulture = frenchCulture;
		}


#if DEBUG
		[DllImport("kernel32.dll", SetLastError = true)]
		[return: MarshalAs(UnmanagedType.Bool)]
		static extern bool AllocConsole();

		void Form1_Load(object sender, EventArgs e)
		{
			AllocConsole();
		}
#endif

		void OnFormClosed(object sender, FormClosedEventArgs e)
		{
			signing?.Dispose();
		}

		Bitmap DocumentPictureBoxResized(Bitmap bitmap)
		{
			lock (documentPictureBoxLock)
			{
				return new Bitmap(bitmap, documentPictureBox.Width, documentPictureBox.Height);
			}
			//return MainFormModel.ResizeHQ(bitmap, documentPictureBox.Width, documentPictureBox.Height);
		}

		public string? SelectDocument(FileDialog fileDialog)
		{
			fileDialog.InitialDirectory = Path.Combine(Environment.CurrentDirectory, "Documents");
			fileDialog.Filter = "pdf documents (*.pdf)|*.pdf";
			fileDialog.FileName = string.Empty;
			DialogResult result = fileDialog.ShowDialog();
			if (DialogResult.OK != result)
			{
				LogToTextBox("Cancelled");
				return null;
			}
			return fileDialog.FileName;
		}

		void SetPageNumberLabel(int number) => PageNumberLabel.Text = $"{number}/{PagesScrollBar.Maximum}";
		void SetPagesScrollBarValue(int pageNumber) => PagesScrollBar.Value = pageNumber;

		async Task SetRenderedPage(int pageNumber)
		{
			Bitmap? bitmap = await MainFormModel.RenderPageAsync(signing, pageNumber, dpi);
			if (bitmap is null) return;
			lock (documentPictureBoxLock)
			{
				documentPictureBox.Image = null;
				documentPictureBox.BackgroundImage = DocumentPictureBoxResized(bitmap);
			}
		}

		async Task SetPageNumberAndRender(int pageNumber)
		{
			SetPagesScrollBarValue(pageNumber);
			SetPageNumberLabel(pageNumber);
			await SetRenderedPage(pageNumber);
		}

		async void OnStartWithDocClicked(object? sender, EventArgs e)
		{
			string? documentPath = SelectDocument(openFileDialog);
			if (documentPath is null) return;
			LogToTextBox($"Selected document: {documentPath}.");

			(string error, ISigning? nextSigning) = await MainFormModel.LoadDocument(documentPath);
			if (nextSigning is null)
			{
				LogToTextBox(error);
				return;
			}
			signing?.Dispose();
			signing = nextSigning;
			PagesScrollBar.Maximum = -1 + await MainFormModel.GetNumberOfPagesAsync(signing);
			await SetPageNumberAndRender(0);
		}

		delegate void SafeCallDelegate(string text, bool withNewLine);
		void LogToTextBox(string? text, bool withNewLine = true)
		{
			if (text is null) return;
			if (infoTextBox.InvokeRequired)
			{
				SafeCallDelegate safeCall = new SafeCallDelegate(LogToTextBox);
				infoTextBox.BeginInvoke(safeCall, new object[] { text, withNewLine });
			}
			else
			{
				infoTextBox.AppendText(text);
				if (withNewLine) infoTextBox.AppendText(Environment.NewLine);
			}
		}

		async void OnStartSaveDocClicked(object? sender, EventArgs e)
		{
			string? documentPath = SelectDocument(saveFileDialog);
			if (documentPath is null)
			{
				LogToTextBox("Something failed with file selection.");
				return;
			}
			LogToTextBox($"Start saving the current document to {documentPath}");
			bool success = await MainFormModel.StartSaveDoc(signing, documentPath);
			if (!success) LogToTextBox("Something failed during file save.");
			else LogToTextBox($"Saved.");
		}


		async void OnScroll(object sender, ScrollEventArgs e)
		{
			switch (e.Type)
			{
				case ScrollEventType.EndScroll:
					await SetPageNumberAndRender(e.NewValue);
					break;
				case ScrollEventType.ThumbTrack:
					if (thumbTrackTask?.IsCompleted ?? true) thumbTrackTask = SetPageNumberAndRender(e.NewValue);
					break;
				default: return;
			}
		}

		int CurrentPageNumber() => PagesScrollBar.Value;

		async Task SetDpi(int dpi)
		{
			LogToTextBox($"Setting dpi = {dpi}");
			this.dpi = dpi;
			await SetRenderedPage(CurrentPageNumber());
		}

		async Task SetDpi300() => await SetDpi(300);

		async Task SetDpi150() => await SetDpi(150);

		async Task SetDpi96() => await SetDpi(96);

		async Task SetDpi72() => await SetDpi(72);

		async void OnDpiItemClicked(object sender, ToolStripItemClickedEventArgs e)
		{
			foreach (ToolStripMenuItem item in menuDpi.DropDownItems)
			{
				item.Checked = false;
			}
			ToolStripMenuItem clickedMenuItem = (ToolStripMenuItem)e.ClickedItem;
			clickedMenuItem.Checked = true;

			if (menuDpi72 == clickedMenuItem) await SetDpi72();
			else if (menuDpi96 == clickedMenuItem) await SetDpi96();
			else if (menuDpi150 == clickedMenuItem) await SetDpi150();
			else if (menuDpi300 == clickedMenuItem) await SetDpi300();
			else return;
		}

		void SetMouseDownPosition(MouseEventArgs e)
		{
			mouseDownPosition = e.Location;
		}

		void SetMouseMovePosition(MouseEventArgs e)
		{
			mouseMovePosition = e.Location;
		}

		void ActivateRectangleDrawing() => isDrawingRectangle = true;

		void DeactivateRectangleDrawing()
		{
			isDrawingRectangle = false;
			menuSignRectangle.Checked = false;
			menuSignInDocument.Checked = false;
			menuViewThenSign.Checked = false;
		}

		void NullExistingPositions()
		{
			mouseMovePosition = null;
			mouseDownPosition = null;
		}

		void OnSignRectangleItemClicked(object? sender, EventArgs e)
		{
			if (isDrawingRectangle || isSigning)
			{
				LogToTextBox("Rectangle drawing or signing should be active already.");
				menuSignRectangle.Checked = true;
				menuSignInDocument.Checked = true;
				menuViewThenSign.Checked = true;
				return;
			}
			NullExistingPositions();
			ActivateRectangleDrawing();
		}

		void OnViewThenSignClicked(object? sender, EventArgs e)
		{
			isSignInDocument = true;
			isViewThenSign = true;
			OnSignRectangleItemClicked(sender, e);
		}

		void OnSignInDocumentClicked(object? sender, EventArgs e)
		{
			isSignInDocument = true;
			OnSignRectangleItemClicked(sender, e);
		}

		bool ShallSigningStart(MouseEventArgs e)
		{
			if (signing is null)
			{
				LogToTextBox("Need a document to sign.");
				return false;
			}
			else return !isDrawingRectangle && drawnRectangle.HasValue;
		}


		string GetSignatureFieldName()
		{
			string dateTimeString = $"{DateTime.UtcNow.ToShortDateString()}_{DateTime.UtcNow.ToLongTimeString()}";
			string signatureFieldName = "SigField_" + dateTimeString;
			LogToTextBox($"Using the field name: {signatureFieldName}");
			return signatureFieldName.Replace('.', '_');
		}

		void RenderPreview()
		{
			lock (documentPictureBoxLock)
			{
				if (drawnRectangle is null) return;
				Bitmap? bitmap = MainFormModel.ReadSignatureImage(signing.Driver, drawnRectangle.Value.Width, drawnRectangle.Value.Height);
				if (bitmap is null) return;
				Bitmap nextPictureBoxImage = new Bitmap(documentPictureBox.Width, documentPictureBox.Height);
				using Graphics g = Graphics.FromImage(nextPictureBoxImage);
				g.DrawRectangle(Pens.Red, drawnRectangle.Value);
				g.DrawImage(bitmap, drawnRectangle.Value.Location);
				documentPictureBox.Image = nextPictureBoxImage;
			}
		}

		void ClearPictureBox()
		{
			lock (documentPictureBoxLock)
			{
				Bitmap nextPictureBoxImage = new Bitmap(documentPictureBox.Width, documentPictureBox.Height);
				documentPictureBox.Image = nextPictureBoxImage;
			}
		}

		async Task SignWithPreview(ISigning signing)
		{
			if (isSigning)
			{
				LogToTextBox("The signature should be active already.");
				return;
			}
			isSigning = true;
			LogToTextBox("Signing active. ", false);

			CancellationTokenSource cancellationTokenSource = new CancellationTokenSource();
			CancellationToken token = cancellationTokenSource.Token;
			Task previewTask = Task.Run(() =>
			{
				while (!token.IsCancellationRequested)
				{
					bool signImgChanged = MainFormModel.WaitForSignImgChanged(signing.Driver, 100);
					if (signImgChanged) RenderPreview();
				}
			}, token);
			LogToTextBox("Please draw.");
			Size documentPictureBoxSize;
			lock (documentPictureBoxLock)
			{
				documentPictureBoxSize = documentPictureBox.Size;
			}
			bool success = await MainFormModel.Sign(signing,
				CurrentPageNumber(), drawnRectangle, documentPictureBoxSize, GetSignatureFieldName(),
				isSignInDocument, isViewThenSign, signerName, signerReason, signerLocation, signerStamps);

			cancellationTokenSource.Cancel();
			if (success)
			{
				LogToTextBox(".", false);
				await previewTask.ConfigureAwait(false);
				drawnRectangle = null;
				LogToTextBox(".", false);
				await SetRenderedPage(CurrentPageNumber());
				LogToTextBox(". ", false);
				LogToTextBox("Signing accomplished.");
			}
			else LogToTextBox("Signing failed or cancelled.");

			isSigning = false;
			isSignInDocument = false;
			isViewThenSign = false;
		}


		void DrawTheRectangle()
		{
			if (mouseDownPosition is null) return;
			if (mouseMovePosition is null) return;

			int row = Math.Min(mouseDownPosition.Value.Y, mouseMovePosition.Value.Y);
			int column = Math.Min(mouseDownPosition.Value.X, mouseMovePosition.Value.X);

			int width = Math.Abs(mouseDownPosition.Value.X - mouseMovePosition.Value.X);
			int height = Math.Abs(mouseDownPosition.Value.Y - mouseMovePosition.Value.Y);

			drawnRectangle = new SysDraw.Rectangle(column, row, width, height);
			if (drawnRectangle is null) return;
			lock (documentPictureBoxLock)
			{
				Bitmap bitmapWithRectangle = new Bitmap(documentPictureBox.Width, documentPictureBox.Height);
				using Graphics g = Graphics.FromImage(bitmapWithRectangle);
				g.DrawRectangle(Pens.Red, drawnRectangle.Value);
				documentPictureBox.Image = bitmapWithRectangle;
			}
		}


		Stack<Button> signControlButtons = new Stack<Button>();

		void AddConfirmRepeatCancelButtons()
		{
			int xStep = 70;
			AddSignControlButton("Confirm", MainFormModel.Confirm, 0 * xStep);
			AddSignControlButton("Repeat", MainFormModel.Repeat, 1 * xStep);
			AddSignControlButton("Cancel", MainFormModel.Cancel, 2 * xStep);
		}

		void AddSignControlButton(string text, Action signControlAction, int offset)
		{
			Button button = new Button();
			button.Text = text;

			button.Location = new Point(drawnRectangle.Value.Left + offset, drawnRectangle.Value.Top);
			this.Controls.Add(button);
			signControlButtons.Push(button);

			Action removeButtons = () => this.BeginInvoke(new Action(() =>
			{
				while (0 < signControlButtons.Count)
				{
					Button next = signControlButtons.Pop();
					Controls.Remove(next);
				}
			}));
			button.Click += (object? sender, EventArgs e) =>
			{
				signControlAction();
				removeButtons();
			};
			MainFormModel.SignFinished += removeButtons;
			MainFormModel.OkClicked += removeButtons;
			MainFormModel.CancelClicked += removeButtons;
			MainFormModel.RepeatClicked += removeButtons;
			button.BringToFront();
		}

		void OnDocPicBoxMouseDown(object sender, MouseEventArgs e) => SetMouseDownPosition(e);

		void OnDocPicBoxMouseMove(object sender, MouseEventArgs e)
		{
			if (!isDrawingRectangle) return;
			SetMouseMovePosition(e);
			DrawTheRectangle();
		}

		void MayAskForName()
		{
			if (menuAskForName.Checked)
			{
				AskForNameDialog askForNameDialog = new AskForNameDialog();
				askForNameDialog.ShowDialog();
				signerName = askForNameDialog.SignerName;
				signerReason = askForNameDialog.SignerReason;
				signerLocation = askForNameDialog.SignerLocation;
				signerStamps = askForNameDialog.SignerStamps;
				menuAskForName.Checked = false;
			}
		}

		void OnDocPicBoxMouseUp(object sender, MouseEventArgs e)
		{
			// Here the signing may start.
			if (!isDrawingRectangle) return;
			DeactivateRectangleDrawing();
			MayAskForName();
			if (ShallSigningStart(e)) Task.Run(() => SignWithPreview(signing));
		}


		async void OnViewCurrentPageClicked(object? sender, EventArgs e)
		{
			if (isSigning)
			{
				LogToTextBox("Ignored this document viewing request because a signing process seems to be active.");
				return;
			}
			await MainFormModel.View(signing, CurrentPageNumber(), false);
		}

		void OnCancelClicked()
		{
			LogToTextBox("Signing cancelled.");
			signing?.Driver.SetCustomerLogoMode();
		}

		void OnOkClicked()
		{
			LogToTextBox("Ok clicked. This cancels the signature process.");
			signing?.Driver.SetCustomerLogoMode();
		}

		void PrintTheDocument(object sender, EventArgs e)
		{
			LogToTextBox("Print depends on PDF service v5.3.0 or higher.");
			MainFormModel.Print(signing);
		}
	}
}