﻿using Sig.DeviceAPI;
using Sig.PdfClient;
using Sig.SignAPI;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using DocSign = Sig.SignAPI.DocSign;
using SysDraw = System.Drawing;

namespace SignDocWithPad
{
	public static class MainFormModel
	{
		const double iicm = 2.54;
		const double serviceDpi = 72.0;
		const double serviceDpcm = serviceDpi / iicm;
		const int oneHour_ms = 60 * 60_000;

		readonly static object readSignatureImageLock = new object();

		internal static Action<string> log = LogToConsole;

		internal static event Action? RepeatClicked;
		internal static event Action? CancelClicked;
		internal static event Action? OkClicked;
		internal static event Action SignFinished = () => { };
		internal static event Action OnSignStarted = () => { };
		static Behaviour behaviour = Behaviour.GetDefault();

		static void LogToConsole(string text)
		{
			try
			{
				Console.WriteLine(text + Environment.NewLine);
			}
			catch
			{
				//
			}
		}

		public static async Task<int> GetNumberOfPagesAsync(ISigning signing)
		{
			return await signing.Client.GetNumberOfPagesAsync();
		}

		static void SetSignRectangleToFullScreen(ISigning signing)
		{
			if (Error.SUCCESS != signing.Driver?.SetSignatureRectangle(0, 0, 0, 0))
				log("Problem ocurred while setting the signature rectangle");
		}

		public static async Task<(string error, ISigning? signing)> LoadDocument(string path)
		{
			byte[] document = File.ReadAllBytes(path);

			Driver? driver = Driver.FromCommunicationType(Sig.DeviceCom.CommunicationType.USB);
			driver.SignFinished += (object o, EventArgs e) => SignFinished();
			driver.SignStarted += (object o, EventArgs e) => OnSignStarted();
			// This changes the timeout. The default is 3000.
			//driver.SignFinishedMillisecondsTimeout = 4000;
			// This disables the timeout. Only clicks stop the signing process.
			//driver.IsSignFinishedEnabled = false;

			Error error = driver.EnumerateDevices(out string[]? deviceNames, out _, FilterDeviceKind.dkStepOver);
			if (deviceNames is null || !deviceNames.Any())
			{
				return ($"No devices found. Driver returned {error}", null);
			}
			driver.SetDevice(deviceNames[0]);
			driver.ReadSignatureImageOptions.BackgroundColor = Color.Transparent;
			driver.ReadSignatureImageOptions.LineColor = Color.DarkBlue;
			driver.ReadSignatureImageOptions.ResizeOptions = ResizeOptions.KeepAspectRatio;
			driver.ReadSignatureImageOptions.Alignment = Alignments.Centered;

			// Enables a some buttons on the pad.
			// Subscribe the driver.ButtonEvent to react to button events.
			EnableOkRepeatCancelButtons(driver);

			driver.ButtonEvent += (object? sender, ButtonEventArgs e) =>
			{
				if (ButtonKind.Repeat == e.ButtonKind) RepeatClicked?.Invoke();
				else if (ButtonKind.Cancel == e.ButtonKind) CancelClicked?.Invoke();
				else if (ButtonKind.Ok == e.ButtonKind) OkClicked?.Invoke();
			};

			ISigning? signing = await SigningFactory.StartAsync(driver, document);
			if (signing is null)
			{
				return ("This should not happen. Check the logs.", null);
			}
			return (string.Empty, signing);
		}


		static void EnableOkRepeatCancelButtons(IDriver driver)
		{
			EnableAndShowOkRepeatCancel(ButtonBarMode.Signature);
			EnableAndShowOkRepeatCancel(ButtonBarMode.DocSign);

			void EnableAndShowOkRepeatCancel(ButtonBarMode mode)
			{
				Error error = driver.GetButtonConfig(out ButtonBar? bar, mode);
				if (Error.SUCCESS == error && bar is not null)
				{
					bar.EnableAndShowButton(ButtonKind.Ok);
					bar.EnableAndShowButton(ButtonKind.Repeat);
					bar.EnableAndShowButton(ButtonKind.Cancel);
					driver.SetButtonConfig(bar);
				}
			}
		}

		static void EnableAndShowButton(this ButtonBar bar, ButtonKind buttonKind)
		{
			if (bar.Configs.ContainsKey(buttonKind))
			{
				bar.Configs[buttonKind].Enabled = true;
				bar.Configs[buttonKind].Visible = true;
			}
		}

		internal static async Task<Bitmap?> RenderFirstPageAsync(ISigning signing) => await RenderPageAsync(signing, 0);

		internal static async Task<Bitmap?> RenderPageAsync(ISigning signing, int pageNumber, int dpi = 72)
		{
			return await signing.Client.RenderPageAsync(pageNumber, dpi);
		}

		private static SignatureInfo? BuildSignatureInfo(string? signer, string? reason, string? location)
		{
			if (string.IsNullOrEmpty(signer)) return null;
			else return new SignatureInfo()
			{
				Name = signer,
				Reason = reason ?? "",
				Location = location ?? "",
			};
		}

		internal static async Task<bool> Sign(
			ISigning signing,
			int pageNumber,
			SysDraw.Rectangle? drawnRectangle,
			Size pictureBoxSize,
			string signFieldName,
			bool docSign,
			bool viewThenSign,
			string? name = null,
			string? reason = null,
			string? location = null,
			bool stamp = false)
		{

			if (viewThenSign)
			{
				Task<bool> viewTask = View(signing, pageNumber, true);
				bool viewingDone = false;
				EventHandler<ButtonEventArgs> startSignButtonHandler = (object? sender, ButtonEventArgs e) =>
				{
					if (ButtonKind.StartSignature == e.ButtonKind)
					{
						viewingDone = true;
					}
				};
				signing.Driver.ButtonEvent += startSignButtonHandler;
				bool good = await viewTask;
				SpinWait.SpinUntil(() => viewingDone, oneHour_ms);
				signing.Driver.ButtonEvent -= startSignButtonHandler;
			}

			if (drawnRectangle is null)
			{
				log("Error, a rectangle is required.");
				return false;
			}

			IEnumerable<IPageDimension>? pageDimensions = await signing.Client.GetPagesAsync();
			IPageDimension? pageDim = pageDimensions?.ElementAtOrDefault(pageNumber);
			if (pageDim is null)
			{
				log("Could not retrieve the page dimension.");
				return false;
			}

			SignatureInfo? signatureInfo = BuildSignatureInfo(name, reason, location);
			behaviour = Behaviour.GetDefault();
			if (docSign)
			{
				// The signature rectangle appears near the lower left corner of the device display
				// and should be at least 50mm wide and 30mm high
				// if the display size allows it.
				behaviour.SignInDocument(DocSign.Rectangle.FromBottomLeft(10, 10, 50, 30));
				// Alternatively, use a zoom level. 
				//behaviour.SignInDocument(DocSign.Rectangle.FromBottomLeft(10, 10, Sig.SignAPI.Zoom.Level.FitWidth));
			}
			if (stamp) behaviour.Stamp();

			if (signing is null) return false;
			SetSignRectangleToFullScreen(signing);

			double column0 = pageDim.Width * drawnRectangle.Value.X / pictureBoxSize.Width;
			double column1 = pageDim.Width * drawnRectangle.Value.Right / pictureBoxSize.Width;

			double row0 = pageDim.Height * drawnRectangle.Value.Y / pictureBoxSize.Height;
			double row1 = pageDim.Height * drawnRectangle.Value.Bottom / pictureBoxSize.Height;

			FieldPosition signFieldPosition = new FieldPosition(row0, column0, row1, column1) { PageNumber = pageNumber };
			log($"Rows: {row0}, {row1}. Height = {row1 - row0}");
			log($"Columns: {column0}, {column1}. Width = {column1 - column0}");

			return await signing.SignAsync(signFieldName, signFieldPosition, signatureInfo, behaviour).ConfigureAwait(false);
		}


		internal static async Task<bool> View(ISigning signing, int pageNumber, bool startSignatureButtonEnabled)
		=> await signing.ViewDocumentAsync(pageNumber, null, startSignatureButtonEnabled);

		internal static async Task<bool> StartSaveDoc(ISigning signing, string? documentPath)
		{
			if (null == documentPath) return false;

			byte[]? document = await signing.Client.DownloadPdfAsync();
			if (document is null)
			{
				log("Could not download document");
				return false;
			}

			return TrySaveDocument(documentPath, document);
		}

		private static bool TrySaveDocument(string? documentPath, byte[]? document)
		{
			if (null == documentPath || null == document) return false;
			try
			{
				File.WriteAllBytes(documentPath, document);
				return true;
			}
			catch
			{
				log("Could not write to file.");
				return false;
			}
		}

		internal static Bitmap? ReadSignatureImage(IDriver driver, int width, int height)
		{
			Error error;
			Bitmap bitmap;
			lock (readSignatureImageLock)
			{
				// Questionable thread safety.
				error = driver.ReadSignatureImage(width, height, out bitmap);
			}
			if (Error.SUCCESS != error || bitmap is null)
			{
				log("Reading the signature image failed.");
				return null;
			}
			return bitmap;
		}

		internal static bool WaitForSignImgChanged(IDriver driver, int milliSecondsTimeout)
		{
			bool signImgChanged = false;
			void SetFlag(object? sender, EventArgs e) { signImgChanged = true; }
			driver.SignImgChanged += SetFlag;
			SpinWait.SpinUntil(() => signImgChanged, milliSecondsTimeout);
			driver.SignFinished -= SetFlag;
			return signImgChanged;
		}

		internal static void Confirm()
		=> behaviour.Click(ButtonKind.Ok);

		internal static void Repeat()
		=> behaviour.Click(ButtonKind.Repeat);

		internal static void Cancel()
		=> behaviour.Click(ButtonKind.Cancel);

		internal static void Print(ISigning? signing)
		{
			signing?.Client?.Print(PrintArgs.Default);
		}
	}
}
