﻿using Sig.DeviceAPI;
using Sig.PdfClient;
using Sig.SignAPI;
using System.Linq;
using System.Collections.Generic;
using System.IO;
using System.Threading.Tasks;
using static System.Console;
using System;
using System.Text.RegularExpressions;

namespace ViewThenSignDocWithPad
{
	class Program
	{

		static async Task Main(string[] arg)
		{
			Driver driver = Driver.FromCommunicationType(Sig.DeviceCom.CommunicationType.USB);

			_ = driver.EnumerateDevices(out string[] devices, out _, FilterDeviceKind.dkStepOver);

			if (devices is null || 0 == devices.Length)
			{
				Error("No device found.");
				return;
			}

			_ = driver.SetDevice(devices[0]);

			// The SignAPI requires a running PDF service.
			// See https://www.stepoverinfo.net/confluence/x/owS-Aw 
			byte[] document = await GetDocument();
			ISigning signing = await SigningFactory.StartAsync(driver, document);
			if (signing is null)
			{
				Error("Failed. Is the PDF service running?");
				return;
			}

			driver.ButtonEvent += OnButtonEvent;
			Task<bool> viewTask = signing.ViewDocumentAsync(0, Sig.SignAPI.Zoom.Level.FitWidth, true);

			Task signTextTask = ExtractAndSetSignText(driver, signing);
			bool viewSuccess = await viewTask;
			if (!viewSuccess)
			{
				Error("Failed. Is the PDF service running?");
				return;
			}
			await signTextTask;

			PrintClickReturnToExit();
			_ = ReadLine();
			driver.SetCustomerLogoMode();
			signing.Dispose();

			async void OnButtonEvent(object sender, ButtonEventArgs e)
			{
				if (ButtonKind.StartSignature == e.ButtonKind)
				{
					driver.ButtonEvent -= OnButtonEvent;
					WriteLine("The start sign button was clicked.");

					IEnumerable<FieldPosition> stuttgartPositions = await signing.Client.FindTextAsync("Stuttgart");
					FieldPosition stuttgart = stuttgartPositions.First();

					double cm = 72 / 2.54;
					double width = 5 * cm;
					double height = 3 * cm;
					double row0 = stuttgart.Row1;
					double column0 = stuttgart.Column0;
					double row1 = row0 + height;
					double column1 = column0 + width;
					FieldPosition fieldPosition = new FieldPosition(row0, column0, row1, column1)
					{
						PageNumber = stuttgart.PageNumber
					};

					string fieldName = "signField_" + DateTime.UtcNow.ToString().Replace('.', '_');
					await signing.SignAsync(fieldName, fieldPosition);

					byte[] signedDocument = await signing.Client.DownloadPdfAsync();
					const string path = "signedDocument.pdf";
					File.WriteAllBytes(path, signedDocument);
					WriteLine($"Saved the signed document to {path}.");
					PrintClickReturnToExit();
				}
			}
		}

		static async Task ExtractAndSetSignText(Driver driver, ISigning signing)
		{
			// Extract the number in Keyword: 123456789.
			// Depends on PDF service v5.3.0 or higher.
			Text text = await signing.Client.ExtractText();
			string textOnFirstPage = text.Pages()[0];
			MatchCollection matches = Regex.Matches(textOnFirstPage, @"Keyword: (\d+)");
			string keyword = "";
			foreach (Match match in matches)
			{
				keyword = match.Groups[1].Value;
				break;
			}
			driver.SetSignText($"Keyword: {keyword}", 32);
		}

		private static void Error(string value)
		{
			WriteLine(value);
			PrintClickReturnToExit();
			_ = ReadLine();
		}

		static async Task<byte[]> GetDocument()
		{
			const string resourcePath = "ViewThenSignDocWithPad.Mustermann_Test_Formfields.pdf";
			Stream stream = typeof(Program).Assembly.GetManifestResourceStream(resourcePath);
			MemoryStream mStream = new MemoryStream();
			await stream.CopyToAsync(mStream);
			byte[] document = mStream.ToArray();
			return document;
		}

		static void PrintClickReturnToExit()
		=> WriteLine("Click return to exit.");
	}
}
